#include "stdafx.h"
#include "clf_primitives.h"
#include "clf_world.h"

clf_vector3D clf_world_object::gravity = clf_vector3D(0.f, -9.81f, 0.f);

clFloat clf_world::sim_speed = .2f;	// used to slow down (like slow motion) the simulation

// base object type
clf_world_object::clf_world_object(clFloat m, clFloat r ):m_mass(m), m_radius(r), m_radius2(r*r)
{
	m_fixed_pos = false;
}

clf_world_object::~clf_world_object()
{
}

void clf_world_object::step_simulation(clFloat fElapsed)
{
	if (m_fixed_pos)
		return;

	m_pos = m_pos+m_vel * fElapsed  + fElapsed*fElapsed*0.5f*gravity;
	m_vel = m_vel+gravity*fElapsed;
}

void clf_world_object::render()
{
}

// only check in the X-Z plane, the ground of our world
bool clf_world_object::is_at_xz(clf_vector3D& pos, clFloat r)
{
	clFloat dx = (pos.x - m_pos.z)-r;
	clFloat dz = (pos.z - m_pos.z)-r;

	clFloat d = dx*dx+dz*dz;
	if (d <= m_radius2)
		return true;

	return false;
}

/////////////////////////////////////////////////////////
// our game world
clf_world::clf_world()
{
}

clf_world::~clf_world()
{
}

// iterate the objects and find collisions
void clf_world::collision_detection()
{
	std::list<clf_world_object*>::iterator it;
	for (it=m_objects.begin(); it!=m_objects.end(); ++it)
	{
		std::list<clf_world_object*>::iterator jt;
		jt = it;
		++jt;
		clf_world_object* iob = *it;
		for ( ; jt!=m_objects.end(); ++jt)
		{
			clf_world_object* job = *jt;
			if (iob->m_fixed_pos && job->m_fixed_pos)
				continue;
			iob->collides_with(job);
		}
	}
}

// calculate next frame
void clf_world::step_simulation(clFloat fElapsed)
{
	std::list<clf_world_object*>::iterator it;
	fElapsed *= sim_speed;
	for (it=m_objects.begin(); it!=m_objects.end(); ++it)
	{
		// move objects
		(*it)->step_simulation(fElapsed);
	}
	// check for collisions
	collision_detection();
}

// draw the objects
void clf_world::render()
{
	std::list<clf_world_object*>::iterator it;
	for (it=m_objects.begin(); it!=m_objects.end(); ++it)
	{
		(*it)->render();
	}
}

// add object to world
void clf_world::add_object(clf_world_object* pob)
{
	m_objects.push_back(pob);
}

clf_world_object* clf_world::object_at(clf_vector3D& pos, clFloat r)
{
	clf_world_object* ob = NULL;

	std::list<clf_world_object*>::iterator it;
	for (it=m_objects.begin(); it!=m_objects.end(); ++it)
	{
		if ((*it)->is_at_xz(pos, r))
			return (*it);
	}

	return ob;
}

// the sphere object
// enable/disable definition to see the simulation with gravity effect or not
#define NO_GRAVITY
clf_sphere::clf_sphere(clFloat m, clFloat r ):clf_world_object(m,r)
{
}

clf_sphere::~clf_sphere()
{
}

void clf_sphere::render()
{
	glPushMatrix();
	glTranslatef(m_pos.x, m_pos.y, m_pos.z);
	drawSphere(m_radius);
	glPopMatrix();
}

void clf_sphere::step_simulation(clFloat fElapsed)
{
	// here we implement trajectory without gravity
#if defined NO_GRAVITY
	m_pos = m_pos+m_vel * fElapsed;
#else
	// base class implements gravity
	clf_world_object::step_simulation(fElapsed);
#endif
}

// calculate the result of the collision with another sphere
void clf_sphere::collision_respond(clf_sphere* pob)
{
	clf_vector3D x(pob->m_pos - m_pos);
	x.normalize();
	clf_vector3D u1x = x * dotProduct(x, m_vel);
	clf_vector3D u1y = m_vel - u1x;
	clf_vector3D u2x = -x * dotProduct(-x, pob->m_vel);
	clf_vector3D u2y = pob->m_vel - u2x;

	clFloat e = .85f;
	clFloat tot_mass = m_mass+pob->m_mass;
	clf_vector3D v1x = ((m_mass-e*pob->m_mass)/tot_mass)*u1x + ((1+e)*pob->m_mass/tot_mass)*u2x;
	clf_vector3D v2x = ((1+e)*m_mass/tot_mass)*u1x + ((pob->m_mass-e*m_mass)/tot_mass)*u2x;

	m_vel = u1y+v1x;
	pob->m_vel = u2y+v2x;

	x *= (m_radius + pob->m_radius);
	x *= 1.0001f;
	pob->m_pos = m_pos + x;
}

// check if the two spheres collide
bool clf_sphere::collides_with(clf_sphere* pob)
{
	clf_vector3D dist_vec(pob->m_pos - m_pos);
	clFloat dist = dist_vec.length();
	if (dist > (m_radius + pob->m_radius)	)
		return false;
	collision_respond(pob);
	return true;
}

// general collision detection function
bool clf_sphere::collides_with(clf_world_object* pob)
{
	clf_sphere* ob = dynamic_cast<clf_sphere*>(pob);
	if (ob)
		return collides_with(ob);
	return true;
}
//////////////////////////////////////////////////////////////
void building::render()
{
	static GLfloat n[6][3] =
	{
		{-1.0, 0.0, 0.0},	// left face     *
		{0.0, 1.0, 0.0},  // top face      *
		{1.0, 0.0, 0.0},  // right face    *
		{0.0, -1.0, 0.0}, // bottom face   *
		{0.0, 0.0, 1.0},  // front face    *
		{0.0, 0.0, -1.0}  // back face     *
	};
	static GLint faces[6][4] =
	{
		{3, 0, 1, 2},	// left face     *
		{6, 7, 3, 2}, // top face      *
		{6, 5, 4, 7}, // right face    *
		{4, 5, 1, 0}, // bottom face   *
		{2, 1, 5, 6}, // front face    *
		{7, 4, 0, 3}  // back face     *
	};

	GLfloat v[8][3];
	GLint i;

	v[0][0] = v[1][0] = v[2][0] = v[3][0] = -m_size.x / 2;
	v[4][0] = v[5][0] = v[6][0] = v[7][0] = m_size.x / 2;

	v[0][1] = v[1][1] = v[4][1] = v[5][1] = -m_size.y / 2;
	v[2][1] = v[3][1] = v[6][1] = v[7][1] = m_size.y / 2;

	v[0][2] = v[3][2] = v[4][2] = v[7][2] = -m_size.z / 2;
	v[1][2] = v[2][2] = v[5][2] = v[6][2] = m_size.z / 2;

	bool tcoords = false;
	if (m_tex >= 0)
	{
		glEnable(GL_TEXTURE_2D);
		glBindTexture(GL_TEXTURE_2D, m_tex);
		tcoords = true;
	}
	glPushMatrix();
	glTranslatef(m_pos.x, m_pos.y, m_pos.z);

	for (i = 5; i >= 0; i--) 
	{
		if (i==1 || i==3) continue;
		glBegin(GL_QUADS);
		glNormal3fv(&n[i][0]);
		if (tcoords) glTexCoord2f(0, 0);
		glVertex3fv(&v[faces[i][2]][0]);
		if (tcoords) glTexCoord2f(0, 1);
		glVertex3fv(&v[faces[i][3]][0]);
		if (tcoords) glTexCoord2f(1, 1);
		glVertex3fv(&v[faces[i][0]][0]);
		if (tcoords) glTexCoord2f(1, 0);
		glVertex3fv(&v[faces[i][1]][0]);
		glEnd();
	}
	if (m_tex >= 0)
	{
		glDisable(GL_TEXTURE_2D);
	}

	i=1;
	glBegin(GL_QUADS);
	glNormal3fv(&n[i][0]);
	glVertex3fv(&v[faces[i][2]][0]);
	glVertex3fv(&v[faces[i][3]][0]);
	glVertex3fv(&v[faces[i][0]][0]);
	glVertex3fv(&v[faces[i][1]][0]);
	glEnd();

	glPopMatrix();
}
